//////////////////////////////////////////////////////////////////////
// sangoma_interface.cpp: interface for Sangoma API driver.
//
// Author	:	David Rokhvarg	<davidr@sangoma.com>
//////////////////////////////////////////////////////////////////////

#include "sangoma_interface.h"

#define DBG_IFACE	if(1)printf
#define INFO_IFACE	if(1)printf
#define ERR_IFACE	if(1)printf

#if defined(OVERLAPPED_IO)
#define INIT_OVERLAPPED_STRUCT(o)\
{	\
	o.Internal = NULL; \
    o.InternalHigh = NULL; \
    o.Offset = 0; \
    o.OffsetHigh = 0;\
}

#define WAIT_FOR_OVERLAPPED_IO(ov)\
{	\
	DWORD OverlappedWaitResult = WaitForSingleObject(ov.hEvent, INFINITE);\
	if(WAIT_OBJECT_0 != OverlappedWaitResult)\
	{	\
		ERR_IFACE("%s(): overlapped wait failed!\n", __FUNCTION__); \
		return 1;	\
	}	\
}

#define HANDLE_DEVICE_IOCTL_RESULT(ov) \
{ \
	if(GetLastError() == ERROR_IO_PENDING || GetLastError() == ERROR_SUCCESS || GetLastError() == ERROR_INVALID_PARAMETER){ \
		WAIT_FOR_OVERLAPPED_IO(ov); \
	}else{ \
		/* check message log */ \
		printf("%s(): Line: %d: Error!!\n", __FUNCTION__, __LINE__); \
		DecodeLastError(__FUNCTION__); \
		return 1; \
	} \
}

#else
#define INIT_OVERLAPPED_STRUCT(o)

#define HANDLE_DEVICE_IOCTL_RESULT(bResult)\
{ \
	if(bResult == 0){ \
		/* check message log */ \
		printf("%s(): Line: %d: Error!!\n", __FUNCTION__, __LINE__); \
		DecodeLastError(__FUNCTION__); \
		return 1; \
	} \
}

#endif

sangoma_interface::sangoma_interface(int wanpipe_number, int interface_number)
{
	DBG_IFACE( "sangoma_interface::sangoma_interface()\n");

	memset(device_name, 0x00, DEV_NAME_LEN);

	//Form the Interface Name from Wanpipe Number and Interface Index (i.e. WANPIPE1_IF0).
	_snprintf(device_name, DEV_NAME_LEN, "\\\\.\\WANPIPE%d_IF%d", 
		wanpipe_number, interface_number); 

	WanpipeNumber = wanpipe_number;
	InterfaceNumber = interface_number;

	INFO_IFACE( "Using Device Name: %s\n", device_name);

	tx_hdr = (api_tx_hdr_t *)wanpipe_api_cmd.data;//do NOT change this pointer!!

	//////////////////////////////////////////////////////////////////
	hGeneralCommands = INVALID_HANDLE_VALUE;
	hRx = INVALID_HANDLE_VALUE;
	hTx = INVALID_HANDLE_VALUE;

	terminate_tx_rx_threads = 0;
	is_rbs_monitoring_enabled = 0;

	//////////////////////////////////////////////////////////////////
	//receive stuff
	rx_frames_count = 0;
	rx_bytes_count = 0;
	//for counting frames with CRC/Abort errors
	bad_rx_frames_count = 0;

	//////////////////////////////////////////////////////////////////
	//transmit stuff
	hTxWaitHandle = NULL;
	tx_bytes_count = 0;
	tx_frames_count = 0;
	tx_test_byte = 0;

	//////////////////////////////////////////////////////////////////
	//IOCTL management structures and variables
	protocol_cb_size = sizeof(wan_mgmt_t)+sizeof(wan_cmd_t)+1;
	wan_protocol = 0;
	adapter_type = 0;
#if USE_STELEPHONY_API
	stelObj = NULL;
#endif
	generate_bit_rev_table();
#if defined(OVERLAPPED_IO)
	INIT_OVERLAPPED_STRUCT(mOverlappedRx);
	INIT_OVERLAPPED_STRUCT(mOverlappedTx);
	INIT_OVERLAPPED_STRUCT(mOverlappedIdleTxBuffer);
	INIT_OVERLAPPED_STRUCT(mOverlappedManagement);
	INIT_OVERLAPPED_STRUCT(mOverlappedTdmvApiIoctl);

	mOverlappedRx.hEvent = CreateEvent( NULL, FALSE, FALSE, NULL);
	mOverlappedTx.hEvent = CreateEvent( NULL, FALSE, FALSE, NULL);
	mOverlappedIdleTxBuffer.hEvent = CreateEvent( NULL, FALSE, FALSE, NULL);
	mOverlappedManagement.hEvent = CreateEvent( NULL, FALSE, FALSE, NULL);
	mOverlappedTdmvApiIoctl.hEvent = CreateEvent( NULL, FALSE, FALSE, NULL);
#endif
}

sangoma_interface::~sangoma_interface()
{
	DBG_IFACE( "sangoma_interface::~sangoma_interface()\n");
	cleanup();
}

int sangoma_interface::init(callback_functions_t *callback_functions_ptr)
{
	DBG_IFACE("sangoma_interface::init()\n");

	memcpy(&callback_functions, callback_functions_ptr, sizeof(callback_functions_t));

	////////////////////////////////////////////////////////////////////////////
	//Open device for general commands
	hGeneralCommands = open_api_device(device_name);
    if (hGeneralCommands == INVALID_HANDLE_VALUE){
		ERR_IFACE( "Unable to open %s for general commands!\n", device_name);
		return 1;
	}

	////////////////////////////////////////////////////////////////////////////
	//open another handle strictly for reading
	hRx = open_api_device(device_name);
    if(hRx == INVALID_HANDLE_VALUE){
		ERR_IFACE( "\nUnable to open %s for reading!\n", device_name);
		return 1;
	}

	////////////////////////////////////////////////////////////////////////////
	//open another handle strictly for transmission
	hTx = open_api_device(device_name);
    if(hTx == INVALID_HANDLE_VALUE){
		ERR_IFACE( "\nUnable to open %s for transmission!\n", device_name);
		return 1;
	}

	////////////////////////////////////////////////////////////////////////////
	//get current protocol
	if(get_wan_config() == WAN_FALSE){
		ERR_IFACE( "Failed to get current protocol!\n");
		return 1;
	}

	////////////////////////////////////////////////////////////////////////////
	//get Front End Type (T1/E1/Analog...)
	if (get_fe_type(&adapter_type) == WAN_FALSE){
		ERR_IFACE( "Failed to get Front End Type!\n");
		return 1;
	}

	////////////////////////////////////////////////////////////////////////////
	//may need interface configuration, so get it now
	if(get_interface_configuration(&wanif_conf_struct)){
		ERR_IFACE( "Failed to get Interface Configuration!\n");
		return 1;
	}

	////////////////////////////////////////////////////////////////////////////
#if USE_STELEPHONY_API
	//Sangoma Telephony API (stelephony.dll) provides the following telephony services:
	//1. FSK Caller ID detection for Analog FXO.
	//2. Software DTMF detection.
	scf.FSKCallerIDEvent = callback_functions.FSKCallerIDEvent;
	scf.DTMFEvent = callback_functions.DTMFEvent;
	//uncomment it only if running on top of D-channel
	//scf.Q931Event = callback_functions.Q931Event;
	StelSetup(&stelObj, this, &scf);
	DBG_IFACE("%s(): stelObj: 0x%p\n", __FUNCTION__, stelObj);
	if(stelObj){
		stelephony_option_t codec = STEL_OPTION_MULAW;
		StelEventControl(stelObj, STEL_EVENT_FSK_CALLER_ID, STEL_CTRL_CODE_ENABLE, &codec);
		StelEventControl(stelObj, STEL_EVENT_DTMF, STEL_CTRL_CODE_ENABLE, &codec);
		//uncomment it only if running on top of D-channel
		//StelEventControl(stelObj, STEL_EVENT_Q931, STEL_CTRL_CODE_ENABLE, NULL);
    }else{
		/* Possible reasons:
		* 1.ToneDecoder.dll was not registered with COM. 
		*	Resolution: in Command Line window run: regsvr32 ToneDecoder.dll.
		*/
		ERR_IFACE("Failed to initialize Stelephony.dll!\n");
		return 1;
	}
#endif

	return 0;
}

HANDLE sangoma_interface::open_api_device(char *api_device_name)
{
	DBG_IFACE("sangoma_interface::open_api_device():%s\n", api_device_name);

#if defined(OVERLAPPED_IO)
	return CreateFile(	api_device_name, 
						GENERIC_READ | GENERIC_WRITE, 
						FILE_SHARE_READ | FILE_SHARE_WRITE,
						(LPSECURITY_ATTRIBUTES)NULL, 
						OPEN_EXISTING,
						FILE_FLAG_NO_BUFFERING | FILE_FLAG_WRITE_THROUGH | FILE_FLAG_OVERLAPPED,
						(HANDLE)NULL
						);
#else
	return CreateFile(	api_device_name, 
						GENERIC_READ | GENERIC_WRITE, 
						FILE_SHARE_READ | FILE_SHARE_WRITE,
						(LPSECURITY_ATTRIBUTES)NULL, 
						OPEN_EXISTING,
						FILE_FLAG_NO_BUFFERING | FILE_FLAG_WRITE_THROUGH,
						(HANDLE)NULL
						);
#endif
}

void sangoma_interface::generate_bit_rev_table(void)
{
	unsigned char util_char;
	unsigned char misc_status_byte;
	int i;

	/* generate the bit-reversing table for all unsigned characters */
	for(util_char = 0;; util_char ++) {
		misc_status_byte = 0;			/* zero the character to be 'built' */
		/* process all 8 bits of the source byte and generate the */
		for(i = 0; i <= 7; i ++) {
		      	/* corresponding 'bit-flipped' character */
			if(util_char & (1 << i)) {
				misc_status_byte |= (1 << (7 - i));
			}
		}
		/* insert the 'bit-flipped' character into the table 
		 * at the appropriate location */
		wp_brt[util_char] = misc_status_byte; 

		/* exit when all unsigned characters have been processed */
		if(util_char == 0xFF) {
			break;
		}
	}
}

void sangoma_interface::bit_swap_a_buffer(unsigned char *data, int len)
{
	int i;
	for (i=0; i < len; i++){
		data[i]=wp_brt[data[i]];
	}
}

int sangoma_interface::get_wan_config()
{
	int err = WAN_TRUE;

	/* Get Protocol type */ 
	wan_udp.wan_udphdr_command	= WAN_GET_PROTOCOL;
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if (wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "Error: Command WAN_GET_PROTOCOL failed! return code: 0x%X",
			wan_udp.wan_udphdr_return_code);
		return WAN_FALSE;
	}

	wan_protocol = wan_udp.wan_udphdr_data[0];

	INFO_IFACE( "Device %s running protocol: %s\n",
		device_name, SDLA_DECODE_PROTOCOL(wan_protocol));
	 
	return err;
} 

int sangoma_interface::get_fe_type(unsigned char* adapter_type)
{
	int err = WAN_TRUE;

	/* Read Adapter Type */
	wan_udp.wan_udphdr_command = WAN_GET_MEDIA_TYPE;
	wan_udp.wan_udphdr_data[0] = WAN_MEDIA_NONE;
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "Error: Command WAN_GET_MEDIA_TYPE failed! return code: 0x%X",
			wan_udp.wan_udphdr_return_code);
		return WAN_FALSE;
	}

	*adapter_type =	get_wan_udphdr_data_byte(0);

	INFO_IFACE( "Front End Type: ");
	switch(*adapter_type)
	{
	case WAN_MEDIA_NONE:
		INFO_IFACE( "Serial");
		break;
	case WAN_MEDIA_T1:
		INFO_IFACE( "T1");
		break;
	case WAN_MEDIA_E1:
		INFO_IFACE( "E1");
		break;
	case WAN_MEDIA_56K:
		INFO_IFACE( "56K");
		break;
	case WAN_MEDIA_FXOFXS:
		INFO_IFACE( "Aanalog");
		break;
	case WAN_MEDIA_BRI:
		INFO_IFACE( "ISDN BRI");
		break;
	case WAN_MEDIA_SERIAL:
		INFO_IFACE("Serial");
		break;
	default:
		INFO_IFACE( "Unknown");
		err = WAN_FALSE;
	}
	INFO_IFACE( "\n");

	return err;
}

//return POINTER to data at offset 'off'
unsigned char* sangoma_interface::get_wan_udphdr_data_ptr(unsigned char off) 
{
	unsigned char *p_data = (unsigned char*)&wan_udp.wan_udphdr_data[0];
	p_data += off;
	return p_data;
}

unsigned char sangoma_interface::set_wan_udphdr_data_byte(unsigned char off, unsigned char data) 
{
	unsigned char *p_data = (unsigned char*)&wan_udp.wan_udphdr_data[0];
	p_data[off] = data;
	return 0;
}

//return DATA at offset 'off'
unsigned char sangoma_interface::get_wan_udphdr_data_byte(unsigned char off) 
{
	unsigned char *p_data = (unsigned char*)&wan_udp.wan_udphdr_data[0];
	return p_data[off];
}

void sangoma_interface::get_te1_56k_stat(void)
{
	unsigned char* data = NULL;
	unsigned long 	alarms = 0x00;
	sdla_te_pmon_t*	pmon = NULL;

	switch(adapter_type)
	{
	case WAN_MEDIA_T1:
	case WAN_MEDIA_E1:
	case WAN_MEDIA_56K:
		;//do nothing
		break;

	default:
		ERR_IFACE( "Command invalid for Adapter Type %d.\n", adapter_type);
		return;
	}

	/* Read T1/E1/56K alarms and T1/E1 performance monitoring counters */
	wan_udp.wan_udphdr_command = WAN_FE_GET_STAT;
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if (wan_udp.wan_udphdr_return_code != 0){
		ERR_IFACE( "Failed to read T1/E1/56K statistics.\n");
		return;
	}

	data = get_wan_udphdr_data_ptr(0);

#if 0
	if (wan_protocol == WANCONFIG_CHDLC){
		data = &wan_udp.wan_udphdr_chdlc_data[0];
	}else{
		data = &wan_udp.wan_udphdr_data[0];
	}
#endif

	alarms = *(unsigned long*)data;
	if (adapter_type == WAN_MEDIA_T1 || adapter_type == WAN_MEDIA_E1){
		INFO_IFACE( "***** %s: %s Alarms *****\n\n",
			device_name, (adapter_type == WAN_MEDIA_T1) ? "T1" : "E1");
		INFO_IFACE( "ALOS:\t%s\t| LOS:\t%s\n", 
				WAN_TE_ALOS_ALARM(alarms), 
				WAN_TE_LOS_ALARM(alarms));
		INFO_IFACE( "RED:\t%s\t| AIS:\t%s\n", 
				WAN_TE_RED_ALARM(alarms), 
				WAN_TE_AIS_ALARM(alarms));
		if (adapter_type == WAN_MEDIA_T1){ 
			INFO_IFACE( "YEL:\t%s\t| OOF:\t%s\n", 
					WAN_TE_YEL_ALARM(alarms), 
					WAN_TE_OOF_ALARM(alarms));
		}else{
			INFO_IFACE( "OOF:\t%s\n", 
					WAN_TE_OOF_ALARM(alarms));
		}
	}else{
		INFO_IFACE( "***** %s: 56K CSU/DSU Alarms *****\n\n\n", device_name);
	 	INFO_IFACE( "In Service:\t\t%s\tData mode idle:\t\t%s\n",
			 	INS_ALARM_56K(alarms), 
			 	DMI_ALARM_56K(alarms));
	
	 	INFO_IFACE( "Zero supp. code:\t%s\tCtrl mode idle:\t\t%s\n",
			 	ZCS_ALARM_56K(alarms), 
			 	CMI_ALARM_56K(alarms));

	 	INFO_IFACE( "Out of service code:\t%s\tOut of frame code:\t%s\n",
			 	OOS_ALARM_56K(alarms), 
			 	OOF_ALARM_56K(alarms));
		
	 	INFO_IFACE( "Valid DSU NL loopback:\t%s\tUnsigned mux code:\t%s\n",
			 	DLP_ALARM_56K(alarms), 
			 	UMC_ALARM_56K(alarms));

	 	INFO_IFACE( "Rx loss of signal:\t%s\t\n",
			 	RLOS_ALARM_56K(alarms)); 
	}

	if (adapter_type != WAN_MEDIA_T1 && adapter_type != WAN_MEDIA_E1){
		return;
	}
	pmon = (sdla_te_pmon_t*)&data[sizeof(unsigned long)];
/*
//FIXME:
	INFO_IFACE( "\n\n***** %s: %s Performance Monitoring Counters *****\n\n",
			device_name, (adapter_type == WAN_MEDIA_T1) ? "T1" : "E1");
	INFO_IFACE( "Framing Bit Error:\t%ld\tLine Code Violation:\t%ld\n", 
			pmon->frm_bit_error,
			pmon->lcv);
	if (adapter_type == WAN_MEDIA_T1){
		INFO_IFACE( "Out of Frame Errors:\t%ld\tBit Errors:\t\t%ld\n", 
				pmon->oof_errors,
				pmon->bit_errors);
	}else{
		INFO_IFACE( "Far End Block Errors:\t%ld\tCRC Errors:\t%ld\n", 
				pmon->far_end_blk_errors,
				pmon->crc_errors);
	}
*/
	return;
}

void sangoma_interface::set_lb_modes(unsigned char type, unsigned char mode)
{
	switch(adapter_type)
	{
	case WAN_MEDIA_T1:
	case WAN_MEDIA_E1:
	case WAN_MEDIA_56K:
		;//do nothing
		break;

	default:
		ERR_IFACE( "Command invalid for Adapter Type %d.\n", adapter_type);
		return;
	}
	
	wan_udp.wan_udphdr_command	= WAN_FE_SET_LB_MODE;
	wan_udp.wan_udphdr_data_len	= 2;

	set_wan_udphdr_data_byte(0,type);
	set_wan_udphdr_data_byte(1,mode);

#if 0
	if (wan_protocol == WANCONFIG_CHDLC){
		wan_udp.wan_udphdr_chdlc_data[0] = type;
		wan_udp.wan_udphdr_chdlc_data[1] = mode;
	}else{
		wan_udp.wan_udphdr_data[0] = type;
		wan_udp.wan_udphdr_data[1] = mode;
	}
#endif
	DO_COMMAND(wan_udp);
	if (wan_udp.wan_udphdr_return_code != 0){
		ERR_IFACE( "Failed to %s line loopback mode. rc: 0x%X\n",
			((mode == WAN_TE1_ACTIVATE_LB) ? "activate" : "deactivate"), 
			wan_udp.wan_udphdr_return_code);
	}else{
		INFO_IFACE( "Line loopback mode is %s!\n",
			(mode == WAN_TE1_ACTIVATE_LB) ? "activated" : "deactivated");
	}
	return;
}

unsigned char sangoma_interface::get_adapter_type()
{
	return adapter_type;
}

unsigned int sangoma_interface::get_sub_media()
{
	return wanif_conf_struct.sub_media;
}

USHORT sangoma_interface::receive(RX_DATA_STRUCT *pRx)
{
	return DoReadCommand(hRx, pRx);
}

int sangoma_interface::get_operational_stats(net_device_stats_t *stats)
{
	wan_udp.wan_udphdr_command = READ_OPERATIONAL_STATS; 
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "Error: command READ_OPERATIONAL_STATS failed!\n");
		return 1;
	}

	memcpy(stats, (net_device_stats_t *)get_wan_udphdr_data_ptr(0), sizeof(net_device_stats_t));

	INFO_IFACE( "**** OPERATIONAL_STATS ****\n");

	INFO_IFACE( "\ttx_dropped\t: %lu\n", stats->tx_dropped);
	INFO_IFACE( "\ttx_packets\t: %lu\n", stats->tx_packets);
	INFO_IFACE( "\ttx_bytes\t: %lu\n", stats->tx_bytes);
	INFO_IFACE( "\n");

	//receive stats
	INFO_IFACE( "\trx_packets\t: %lu\n", stats->rx_packets);
	INFO_IFACE( "\trx_bytes\t: %lu\n", stats->rx_bytes);

	INFO_IFACE( "\trx_errors\t: %lu\n", stats->rx_errors);
	INFO_IFACE( "\trx_frame_errors\t: %lu\n", stats->rx_frame_errors);

	INFO_IFACE( "\trx_packets_discarded_rx_q_full\t: %lu\n",
		stats->rx_packets_discarded_rx_q_full);
	INFO_IFACE( "\trx_packets_discarded_excessive_length\t: %lu\n",
		stats->rx_packets_discarded_excessive_length);
	
	INFO_IFACE( "\trx_packets_discarded_too_short\t: %lu\n", 
		stats->rx_packets_discarded_too_short);
	INFO_IFACE( "\n");

	//hardware level stats
	//AFT card
	INFO_IFACE( "\trx_fifo_errors\t: %lu\n", stats->rx_fifo_errors);
	INFO_IFACE( "\ttx_fifo_errors\t: %lu\n", stats->tx_fifo_errors);
	//S508, S514
	INFO_IFACE( "\tnumber_of_times_rx_buffers_out_of_synch_with_card\t: %lu\n",
		stats->number_of_times_rx_buffers_out_of_synch_with_card);

	//BitStream mode only - counter of data blocks transmitted automatically,
	//because no User Data was available. Transmit underrun counter from user
	//point of view.
	INFO_IFACE( "\ttx_idle_data\t: %lu\n", stats->tx_idle_data);
	return 0;
}

int sangoma_interface::flush_operational_stats()
{
	wan_udp.wan_udphdr_command = FLUSH_OPERATIONAL_STATS; 
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "Error: command FLUSH_OPERATIONAL_STATS failed!\n");
		return 1;
	}

	INFO_IFACE( "Command FLUSH_OPERATIONAL_STATS was successful.\n");
	return 0;
}

int sangoma_interface::flush_tx_buffers()
{
	wan_udp.wan_udphdr_command = FLUSH_TX_BUFFERS; 
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "Error: command FLUSH_TX_BUFFERS failed!\n");
		return 1;
	}

	INFO_IFACE( "Command FLUSH_TX_BUFFERS was successful.\n");
	return 0;
}

int sangoma_interface::enable_rbs_monitoring()
{
	int rc;
	if(is_rbs_monitoring_enabled == 0){
		/* enable RBS monitoring one time per card */
		if(sangoma_tdm_enable_rbs_events(20)){
			ERR_IFACE("sangoma_tdm_enable_rbs_events() failed!\n");
			rc = 1;
		}else{
			is_rbs_monitoring_enabled = 1;
			INFO_IFACE("RBS Monitoring successfully enabled.\n");
			rc = 0;
		}
	}else{
		INFO_IFACE("RBS Monitoring already enabled!! Should be done only once.\n");
		rc = 0;
	}
	return rc;
}

void sangoma_interface::set_idle_tx_data_buffer(unsigned char idle_tx_test_byte)
{
	//set data in Idle Tx buffer in the driver
	tx_data.api_header.data_length = 512;//should be the same as MTU/MRU
	tx_data.api_header.operation_status = SANG_STATUS_TX_TIMEOUT;
	//set the actual data for transmission
	memset(tx_data.data, idle_tx_test_byte, tx_data.api_header.data_length);

	if(DoSetIdleTxBufferCommand(hGeneralCommands, &tx_data)){
		//error
		ERR_IFACE("DoSetIdleTxBufferCommand() failed!! Check messages log.\n");
		return;
	}

	if(tx_data.api_header.operation_status != SANG_STATUS_SUCCESS){
		//error
		ERR_IFACE("DoSetIdleTxBufferCommand() failed!!.\n");
		return;
	}
}

//get RBS being received
char sangoma_interface::get_rbs(rbs_management_t *rbs_management_ptr)
{
	wan_udp.wan_udphdr_command = WAN_GET_RBS_BITS;
	memcpy(&wan_udp.wan_udphdr_data[0], rbs_management_ptr, sizeof(rbs_management_t));
	wan_udp.wan_udphdr_data_len = sizeof(rbs_management_t);

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "Error: command WAN_GET_RBS_BITS failed!\n");
		return 1;
	}

	memcpy(rbs_management_ptr, (rbs_management_t *)&wan_udp.wan_udphdr_data[0], 
															sizeof(rbs_management_t));

	INFO_IFACE( "**** WAN_GET_RBS_BITS OK ****\n");

	//INFO_IFACE( "ABCD_bits (HEX): 0x%02X\n", rbs_management_ptr->ABCD_bits);

	INFO_IFACE( "Channel: %d, RX RBS: A:%1d B:%1d C:%1d D:%1d\n",
			rbs_management_ptr->channel, 
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_A) ? 1 : 0,
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_B) ? 1 : 0,
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_C) ? 1 : 0,
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_D) ? 1 : 0);
	return 0;
}

//set RBS to be transmitted
char sangoma_interface::set_rbs(rbs_management_t *rbs_management_ptr)
{
	wan_udp.wan_udphdr_command = WAN_SET_RBS_BITS;
	memcpy(&wan_udp.wan_udphdr_data[0], rbs_management_ptr, sizeof(rbs_management_t));
	wan_udp.wan_udphdr_data_len = sizeof(rbs_management_t);

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "Error: command WAN_SET_RBS_BITS failed!\n");
		return 1;
	}

	INFO_IFACE( "**** WAN_SET_RBS_BITS OK ****\n");

	//INFO_IFACE( "ABCD_bits (HEX): 0x%02X\n", rbs_management_ptr->ABCD_bits);

	INFO_IFACE( "Channel: %d, TX RBS: A:%1d B:%1d C:%1d D:%1d\n",
			rbs_management_ptr->channel, 
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_A) ? 1 : 0,
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_B) ? 1 : 0,
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_C) ? 1 : 0,
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_D) ? 1 : 0);
	return 0;
}

int sangoma_interface::get_interface_configuration(if_cfg_t *wanif_conf_ptr)
{
	wan_udp.wan_udphdr_command = READ_CONFIGURATION;
	wan_udp.wan_udphdr_data_len = sizeof(if_cfg_t);

	memset(wanif_conf_ptr, 0x00, sizeof(if_cfg_t));

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "Error: command READ_CONFIGURATION failed!\n");
		return 1;
	}
	memcpy(wanif_conf_ptr, get_wan_udphdr_data_ptr(0), sizeof(if_cfg_t));

	INFO_IFACE( "**** READ_CONFIGURATION ****\n");
	INFO_IFACE( "Operational Mode\t: %s\n", wanif_conf_ptr->usedby);
	INFO_IFACE( "Active Channels \t: 0x%08X\n", wanif_conf_ptr->active_ch);
	INFO_IFACE( "Interface Number\t: %u\n", wanif_conf_ptr->interface_number);
	INFO_IFACE( "Media type\t\t: %u\n", wanif_conf_ptr->media);
	INFO_IFACE( "Line Mode\t\t: %s\n", wanif_conf_ptr->line_mode);
	switch(wanif_conf_ptr->media)
	{
	case WAN_MEDIA_FXOFXS:
		switch(wanif_conf_ptr->sub_media)
		{
		case MOD_TYPE_FXS:
			INFO_IFACE( "Media sub-type\t\t: FXS\n");
			break;
		case MOD_TYPE_FXO:
			INFO_IFACE( "Media sub-type\t\t: FXO\n");
			break;
		default:
			INFO_IFACE( "Media sub-type\t\t: Unknown!!\n");
			break;
		}
		break;/* WAN_MEDIA_FXOFXS */

	case WAN_MEDIA_BRI: 
		switch(wanif_conf_ptr->sub_media)
		{
		case MOD_TYPE_NT:
			INFO_IFACE( "Media sub-type\t\t: ISDN BRI NT\n");
			break;
		case MOD_TYPE_TE:
			INFO_IFACE( "Media sub-type\t\t: ISDN BRI TE\n");
			break;
		default:
			INFO_IFACE( "Media sub-type\t\t: Unknown!!\n");
			break;
		}
		break;/* WAN_MEDIA_BRI */

	case WAN_MEDIA_SERIAL:
		INFO_IFACE("Media sub-type\t\t: %s\n", INT_DECODE(wanif_conf_ptr->sub_media));
		break;
	}
	INFO_IFACE( "****************************\n");

	return 0;
}

void sangoma_interface::get_api_driver_version (PDRIVER_VERSION version)
{
	wan_udp.wan_udphdr_command = READ_CODE_VERSION; 
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE("Error: command READ_CODE_VERSION failed!\n");
		return;
	}

	memcpy(version, (PDRIVER_VERSION)get_wan_udphdr_data_ptr(0), sizeof(DRIVER_VERSION));

	INFO_IFACE("\nAPI version\t: %d,%d,%d,%d\n",
		version->major, version->minor, version->minor1, version->minor2);
	INFO_IFACE("\n");

	INFO_IFACE("Command READ_CODE_VERSION was successful.\n");
}

void sangoma_interface::get_card_customer_id(u8 *customer_id)
{
	wan_udp.wan_udphdr_command = SIOC_AFT_CUSTOMER_ID; 
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE("Error: command SIOC_AFT_CUSTOMER_ID failed!\n");
		return;
	}

	*customer_id = get_wan_udphdr_data_byte(0);

	INFO_IFACE("\nCard Customer ID\t: 0x%02X\n", *customer_id);

	INFO_IFACE("Command SIOC_AFT_CUSTOMER_ID was successful.\n");
}

int sangoma_interface::get_open_handles_counter()
{
	wan_udp.wan_udphdr_command = GET_OPEN_HANDLES_COUNTER; 
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "Error: command GET_OPEN_HANDLES_COUNTER failed!\n");
		return -1;
	}

	return *(int*)&wan_udp.wan_udphdr_data[0];
}


//It is very important to close ALL open
//handles, because the API will continue
//receive data until the LAST handel is closed.
void sangoma_interface::cleanup()
{
	INFO_IFACE("sangoma_interface::cleanup()\n");

	if(hGeneralCommands != INVALID_HANDLE_VALUE){
		int hndls_counter = get_open_handles_counter();
		if(hndls_counter > 0){
			INFO_IFACE( "There are %d API Handles to be closed.\n", hndls_counter);
		}

		if(is_rbs_monitoring_enabled == 1){
			/* disable RBS monitoring one time per card */
			is_rbs_monitoring_enabled = 0;
			if(sangoma_tdm_disable_rbs_events()){
				ERR_IFACE("sangoma_tdm_disable_rbs_events() failed!\n");
			}
		}


		INFO_IFACE( "Closing hGeneralCommands.\n");
		CloseHandle(hGeneralCommands);
		hGeneralCommands = INVALID_HANDLE_VALUE;
	}

	if(hRx != INVALID_HANDLE_VALUE){
		INFO_IFACE( "Closing hRx.\n");
		CloseHandle(hRx);
		hRx = INVALID_HANDLE_VALUE;
	}

	if(hTx != INVALID_HANDLE_VALUE){
		INFO_IFACE( "Closing hTx.\n");
		CloseHandle(hTx);
		hTx = INVALID_HANDLE_VALUE;
	}

	if(hTxWaitHandle != NULL){
		CloseHandle(hTxWaitHandle);
		hTxWaitHandle = NULL;
	}

#if USE_STELEPHONY_API
	if(stelObj){
		StelCleanup(stelObj);
		stelObj = NULL;
	}
#endif
}

// Blocking write command.
UCHAR sangoma_interface::DoWriteCommand(HANDLE drv, TX_DATA_STRUCT * pTx)
{
	DWORD ln, bIoResult;

	INIT_OVERLAPPED_STRUCT(mOverlappedTx);

	bIoResult = DeviceIoControl(
			drv,
			IoctlWriteCommand,
			(LPVOID)pTx,
			(ULONG)sizeof(TX_DATA_STRUCT),
			(LPVOID)pTx,
			sizeof(TX_DATA_STRUCT),
			(LPDWORD)(&ln),
#if defined(OVERLAPPED_IO)
			&mOverlappedTx
#else
			(LPOVERLAPPED)NULL
#endif
						);

#if defined(OVERLAPPED_IO)
	HANDLE_DEVICE_IOCTL_RESULT(mOverlappedTx);
#else
	HANDLE_DEVICE_IOCTL_RESULT(bIoResult);
#endif
	return 0;
}

// Blocking read command.
USHORT sangoma_interface::DoReadCommand(HANDLE drv, RX_DATA_STRUCT * pRx)
{
	DWORD ln, bIoResult;
	
	INIT_OVERLAPPED_STRUCT(mOverlappedRx);

	bIoResult = DeviceIoControl(
			drv,
			IoctlReadCommand,
			(LPVOID)NULL,
			0L,
			(LPVOID)pRx,
			sizeof(RX_DATA_STRUCT),
			(LPDWORD)(&ln),
#if defined(OVERLAPPED_IO)
			&mOverlappedRx
#else
			(LPOVERLAPPED)NULL
#endif
						);

#if defined(OVERLAPPED_IO)
	HANDLE_DEVICE_IOCTL_RESULT(mOverlappedRx);
#else
	HANDLE_DEVICE_IOCTL_RESULT(bIoResult);
#endif
	return 0;
}

// set data in Idle Transmit buffer (BitStream only).
UCHAR sangoma_interface::DoSetIdleTxBufferCommand(HANDLE drv, TX_DATA_STRUCT * pTx)
{
	DWORD ln, bIoResult;
	
	INIT_OVERLAPPED_STRUCT(mOverlappedIdleTxBuffer);

	bIoResult = DeviceIoControl(
			drv,
			IoctlSetIdleTxBuffer,
			(LPVOID)pTx,
			(ULONG)sizeof(TX_DATA_STRUCT),
			(LPVOID)pTx,
			sizeof(TX_DATA_STRUCT),
			(LPDWORD)(&ln),
#if defined(OVERLAPPED_IO)
			&mOverlappedIdleTxBuffer
#else
			(LPOVERLAPPED)NULL
#endif
						);

#if defined(OVERLAPPED_IO)
	HANDLE_DEVICE_IOCTL_RESULT(mOverlappedIdleTxBuffer);
#else
	HANDLE_DEVICE_IOCTL_RESULT(bIoResult);
#endif
	return 0;
}

int sangoma_interface::DoManagementCommand(HANDLE drv, wan_udp_hdr_t* wan_udp)
{
	DWORD ln, bIoResult;
    static unsigned char id = 0;
	int data_len = protocol_cb_size+wan_udp->wan_udphdr_data_len;

	wan_udp->wan_udphdr_request_reply = 0x01;
	wan_udp->wan_udphdr_id = id;
   	wan_udp->wan_udphdr_return_code = WAN_UDP_TIMEOUT_CMD;

	INIT_OVERLAPPED_STRUCT(mOverlappedManagement);

	bIoResult = DeviceIoControl(
			drv,
			IoctlManagementCommand,
			(LPVOID)wan_udp,
			sizeof(wan_udp_hdr_t),
			(LPVOID)wan_udp,
			sizeof(wan_udp_hdr_t),
			(LPDWORD)(&ln),
#if defined(OVERLAPPED_IO)
			&mOverlappedManagement
#else
			(LPOVERLAPPED)NULL
#endif
						);

#if defined(OVERLAPPED_IO)
	HANDLE_DEVICE_IOCTL_RESULT(mOverlappedManagement);
#else
	HANDLE_DEVICE_IOCTL_RESULT(bIoResult);
#endif
	return 0;
}

int sangoma_interface::tdmv_api_ioctl(wanpipe_tdm_api_cmd_t *api_cmd)
{
	DWORD			ln, bIoResult;
    unsigned char	id = 0;
	int				retry_counter = 0;

	INIT_OVERLAPPED_STRUCT(mOverlappedTdmvApiIoctl);

retry:
	DBG_IFACE( "%s:tdmv_api_ioctl()\n", device_name);

	wan_udp.wan_udphdr_request_reply = 0x01;
	wan_udp.wan_udphdr_id = id;
   	wan_udp.wan_udphdr_return_code = WAN_UDP_TIMEOUT_CMD;

	wan_udp.wan_udphdr_command	= WAN_TDMV_API_IOCTL;
	wan_udp.wan_udphdr_data_len	= sizeof(wanpipe_tdm_api_cmd_t);

	//copy data from caller's buffer to driver's buffer
	memcpy(	wan_udp.wan_udphdr_data, 
			(void*)api_cmd,
			sizeof(wanpipe_tdm_api_cmd_t));

	bIoResult = DeviceIoControl(
			hGeneralCommands,
			IoctlManagementCommand,
			(LPVOID)&wan_udp,
			sizeof(wan_udp_hdr_t),
			(LPVOID)&wan_udp,
			sizeof(wan_udp_hdr_t),
			(LPDWORD)(&ln),
#if defined(OVERLAPPED_IO)
			&mOverlappedTdmvApiIoctl
#else
			(LPOVERLAPPED)NULL
#endif
			);

#if defined(OVERLAPPED_IO)
	HANDLE_DEVICE_IOCTL_RESULT(mOverlappedTdmvApiIoctl);
#else
	HANDLE_DEVICE_IOCTL_RESULT(bIoResult);
#endif

	switch(wan_udp.wan_udphdr_return_code)
	{
	case 0:
		//OK
		break;

	case SANG_STATUS_DEVICE_BUSY:
		retry_counter++;
		if(retry_counter < MAX_RETRY_COUNTER){
			RETRY_DELAY();
			DBG_IFACE( "%s:tdmv_api_ioctl(): Retrying...\n", device_name);
			goto retry;
		}
		break;
	}

	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "tdmv_api_ioctl(): command failed! Return code: %s (0x%X).\n",
			SDLA_DECODE_SANG_STATUS(wan_udp.wan_udphdr_return_code), 
			wan_udp.wan_udphdr_return_code);
		
		return wan_udp.wan_udphdr_return_code;
	}

/*	wanpipe_tdm_api_cmd_t *wanpipe_tdm_api_cmd_ptr = (wanpipe_tdm_api_cmd_t *)wan_udp.wan_udphdr_data;
	printf("%s(): wanpipe_tdm_api_cmd_ptr->fe_status: %d\n",	__FUNCTION__, wanpipe_tdm_api_cmd_ptr->fe_status); */

	//copy data from driver's buffer to caller's buffer
	memcpy(	(void*)api_cmd,
			wan_udp.wan_udphdr_data, 
			sizeof(wanpipe_tdm_api_cmd_t));

	return 0;
}

int sangoma_interface::sangoma_tdm_enable_ring_detect_events() 
{
	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_EVENT;
	tdm_api_cmd.event.wp_tdm_api_event_type = WP_TDMAPI_EVENT_RING_DETECT;
	tdm_api_cmd.event.wp_tdm_api_event_mode = WP_TDMAPI_EVENT_ENABLE;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

int sangoma_interface::sangoma_tdm_disable_ring_detect_events() 
{
	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_EVENT;
	tdm_api_cmd.event.wp_tdm_api_event_type = WP_TDMAPI_EVENT_RING_DETECT;
	tdm_api_cmd.event.wp_tdm_api_event_mode = WP_TDMAPI_EVENT_DISABLE;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

int sangoma_interface::sangoma_tdm_enable_ring_trip_detect_events() 
{
	DBG_IFACE("%s()\n", __FUNCTION__);

	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_EVENT;
	tdm_api_cmd.event.wp_tdm_api_event_type = WP_TDMAPI_EVENT_RING_TRIP_DETECT;
	tdm_api_cmd.event.wp_tdm_api_event_mode = WP_TDMAPI_EVENT_ENABLE;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

int sangoma_interface::sangoma_tdm_disable_ring_trip_detect_events() 
{
	DBG_IFACE("%s()\n", __FUNCTION__);

	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_EVENT;
	tdm_api_cmd.event.wp_tdm_api_event_type = WP_TDMAPI_EVENT_RING_DETECT;
	tdm_api_cmd.event.wp_tdm_api_event_mode = WP_TDMAPI_EVENT_DISABLE;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

int sangoma_interface::sangoma_tdm_enable_rm_dtmf_events() 
{
	DBG_IFACE("%s()\n", __FUNCTION__);

	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_EVENT;
	tdm_api_cmd.event.wp_tdm_api_event_type = WP_TDMAPI_EVENT_RM_DTMF;
	tdm_api_cmd.event.wp_tdm_api_event_mode = WP_TDMAPI_EVENT_ENABLE;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

int sangoma_interface::sangoma_tdm_disable_rm_dtmf_events() 
{
	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_EVENT;
	tdm_api_cmd.event.wp_tdm_api_event_type = WP_TDMAPI_EVENT_RM_DTMF;
	tdm_api_cmd.event.wp_tdm_api_event_mode = WP_TDMAPI_EVENT_DISABLE;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

int sangoma_interface::sangoma_tdm_enable_dtmf_events(int channel)
{
	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_EVENT;
	tdm_api_cmd.event.wp_tdm_api_event_type = WP_TDMAPI_EVENT_DTMF;
	tdm_api_cmd.event.wp_tdm_api_event_mode = WP_TDMAPI_EVENT_ENABLE;
	tdm_api_cmd.event.channel = channel;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

int sangoma_interface::sangoma_tdm_disable_dtmf_events(int channel) 
{
	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_EVENT;
	tdm_api_cmd.event.wp_tdm_api_event_type = WP_TDMAPI_EVENT_DTMF;
	tdm_api_cmd.event.wp_tdm_api_event_mode = WP_TDMAPI_EVENT_DISABLE;
	tdm_api_cmd.event.channel = channel;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

int sangoma_interface::sangoma_tdm_enable_rxhook_events() 
{
	DBG_IFACE("%s()\n", __FUNCTION__);

	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_EVENT;
	tdm_api_cmd.event.wp_tdm_api_event_type = WP_TDMAPI_EVENT_RXHOOK;
	tdm_api_cmd.event.wp_tdm_api_event_mode = WP_TDMAPI_EVENT_ENABLE;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

int sangoma_interface::sangoma_tdm_disable_rxhook_events() 
{
	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_EVENT;
	tdm_api_cmd.event.wp_tdm_api_event_type = WP_TDMAPI_EVENT_RXHOOK;
	tdm_api_cmd.event.wp_tdm_api_event_mode = WP_TDMAPI_EVENT_DISABLE;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

int sangoma_interface::sangoma_tdm_enable_ring_events()
{
	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_EVENT;
	tdm_api_cmd.event.wp_tdm_api_event_type = WP_TDMAPI_EVENT_RING;
	tdm_api_cmd.event.wp_tdm_api_event_mode = WP_TDMAPI_EVENT_ENABLE;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

int sangoma_interface::sangoma_tdm_disable_ring_events() 
{
	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_EVENT;
	tdm_api_cmd.event.wp_tdm_api_event_type = WP_TDMAPI_EVENT_RING;
	tdm_api_cmd.event.wp_tdm_api_event_mode = WP_TDMAPI_EVENT_DISABLE;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

int sangoma_interface::sangoma_tdm_txsig_onhook() 
{
	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_EVENT;
	tdm_api_cmd.event.wp_tdm_api_event_type = WP_TDMAPI_EVENT_TXSIG_ONHOOK;
	tdm_api_cmd.event.wp_tdm_api_event_mode = WP_TDMAPI_EVENT_ENABLE;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

int sangoma_interface::sangoma_tdm_txsig_offhook() 
{
	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_EVENT;
	tdm_api_cmd.event.wp_tdm_api_event_type = WP_TDMAPI_EVENT_TXSIG_OFFHOOK;
	tdm_api_cmd.event.wp_tdm_api_event_mode = WP_TDMAPI_EVENT_ENABLE;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

int sangoma_interface::sangoma_tdm_enable_tone_events(int tone_id) 
{
	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_EVENT;
	tdm_api_cmd.event.wp_tdm_api_event_type = WP_TDMAPI_EVENT_TONE;
	tdm_api_cmd.event.wp_tdm_api_event_mode = WP_TDMAPI_EVENT_ENABLE;
	tdm_api_cmd.event.wp_tdm_api_event_tone_type = tone_id;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

int sangoma_interface::sangoma_tdm_disable_tone_events() 
{
	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_EVENT;
	tdm_api_cmd.event.wp_tdm_api_event_type = WP_TDMAPI_EVENT_TONE;
	tdm_api_cmd.event.wp_tdm_api_event_mode = WP_TDMAPI_EVENT_DISABLE;
	tdm_api_cmd.event.wp_tdm_api_event_tone_type = 0x00;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

/* 1. Enable 'writing' of RBS bits on card.
   2. Enable monitoring change in state of RBS bits.
      'polls_per_second' - how many times per second API driver will check for
      change in state of RBS bits.
      Valid values are between 20 and 100 (including). */
int sangoma_interface::sangoma_tdm_enable_rbs_events(int polls_per_second) 
{
	tdm_api_cmd.cmd = SIOC_WP_TDM_ENABLE_RBS_EVENTS;
	tdm_api_cmd.rbs_poll = polls_per_second; 
	return tdmv_api_ioctl(&tdm_api_cmd);	
}

/* Stop monitoring change in state of RBS bits */
int sangoma_interface::sangoma_tdm_disable_rbs_events() 
{
	tdm_api_cmd.cmd = SIOC_WP_TDM_DISABLE_RBS_EVENTS;
	return tdmv_api_ioctl(&tdm_api_cmd);	
}

/* Activate ISDN BRI line. */
int sangoma_interface::sangoma_tdm_front_end_activate()
{
	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_FE_STATUS;
	tdm_api_cmd.fe_status = WAN_FE_CONNECTED;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

/* De-activate ISDN BRI line. */
int sangoma_interface::sangoma_tdm_front_end_deactivate()
{
	tdm_api_cmd.cmd = SIOC_WP_TDM_SET_FE_STATUS;
	tdm_api_cmd.fe_status = WAN_FE_DISCONNECTED;
	return tdmv_api_ioctl(&tdm_api_cmd);
}

/* get current state of the line - is it Connected or Disconnected */
int sangoma_interface::sangoma_tdm_get_front_end_status(unsigned char *status)
{
	int rc;
	tdm_api_cmd.cmd = SIOC_WP_TDM_GET_FE_STATUS;
	rc = tdmv_api_ioctl(&tdm_api_cmd);
	printf("rc: %d\n", rc);
	if(rc == 0){
		*status = tdm_api_cmd.fe_status;
	}
	return rc;
}

int sangoma_interface::stop()
{
	int wait_counter = 0;
	DBG_IFACE("sangoma_interface::stop()\n");

	INFO_IFACE( "Stopping.");

	terminate_tx_rx_threads = 1;
	while(terminate_tx_rx_threads == 1 && wait_counter++ < 5){
		INFO_IFACE( ".");
		Sleep(500);
	}
	INFO_IFACE( "\n");

	switch(wanif_conf_struct.media)
	{
	case WAN_MEDIA_FXOFXS:
		switch(wanif_conf_struct.sub_media)
		{
		case MOD_TYPE_FXS:
			stop_all_tones();
			stop_ringing_phone();
			break;

		case MOD_TYPE_FXO:
			fxo_go_on_hook();
			break;
		}
		break;
	}

	return 0;
}

int sangoma_interface::run()
{
	DBG_IFACE("sangoma_interface::run()\n");

	switch(wanif_conf_struct.media)
	{
	case WAN_MEDIA_FXOFXS:
		switch(wanif_conf_struct.sub_media)
		{
		case MOD_TYPE_FXS:
			sangoma_tdm_enable_rxhook_events();
			sangoma_tdm_enable_rm_dtmf_events();
			sangoma_tdm_enable_ring_trip_detect_events();
			break;
		case MOD_TYPE_FXO:
			sangoma_tdm_enable_ring_detect_events();
			break;
		}
		break;
	}

	////////////////////////////////////////////////////////////////////////////
	//Start a thread for receiving data only
    if(this->CreateThread(1) == false){
		ERR_IFACE( "Failed to create Rx thread!!\n");
		return 1;
	}

	return 0;
}

unsigned long sangoma_interface::threadFunction (void *parameter)
{
	switch((int)parameter)
	{
	case 1:
		RxThreadFunc();
		return 0;//ok
	}
	return 1;//invalid thread function was requested
}

// Read Thread
void sangoma_interface::RxThreadFunc()
{ 
	INFO_IFACE( "\n%s: RxThreadFunc() - start. ThreadID: %d\n", device_name, ::GetCurrentThreadId()); 

	if(SetThreadPriority(::GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL) == FALSE){
		ERR_IFACE( "Failed to set Rx thread priority!!\n");
		return;
	}

	while(terminate_tx_rx_threads == 0){	
		if(read_data()){
			break;
		}
	}//while()

	terminate_tx_rx_threads = 2;

    INFO_IFACE( "\n%s: RxThreadFunc() - end\n", device_name); 
}

///////////////////////////////////////////////////////////////////////

int sangoma_interface::read_data()
{
	api_header_t		*pri;

	if(receive(&rx_data) ){
		//error
		ERR_IFACE( "receive() failed!! Check messages log.\n");
		return 1;
	}

	//use our special buffer at rxdata to hold received data
	pri = &rx_data.api_header;

	switch(pri->operation_status)
	{
	case SANG_STATUS_RX_DATA_TIMEOUT:
		//No data in READ_CMD_TIMEOUT, try again. (NOT and error)
		DBG_IFACE("Timeout on read.\n");
		break;

	case SANG_STATUS_BUFFER_TOO_SMALL:
		//Recieved data longer than the pre-configured maximum.
		//Maximum length is set in 'Interface Properties',
		//in the 'Device Manager'.
		ERR_IFACE( "Received data longer than buffer passed to API.\n");
		break;

	case SANG_STATUS_LINE_DISCONNECTED:
		INFO_IFACE( "%s: SANG_STATUS_LINE_DISCONNECTED\n", device_name);
		break;

	case SANG_STATUS_LINE_CONNECTED:
		INFO_IFACE( "%s: SANG_STATUS_LINE_CONNECTED\n", device_name);
		break;

	case SANG_STATUS_RX_DATA_AVAILABLE:
#if USE_STELEPHONY_API
		if(stelObj){
			RX_DATA_STRUCT* pRx = (TX_RX_DATA_STRUCT*)&rx_data;
			StelStreamInput(stelObj, pRx->data, pRx->api_header.data_length);
		}
#endif

#if 0
		//Some useful information about API's internal receive queue
		//is available after each successful IoctlReadCommand:
		DBG_IFACE("max_rx_queue_length: %d current_number_of_frames_in_rx_queue: %d\n",  
			rx_data.api_header.rx_h.max_rx_queue_length,
			rx_data.api_header.rx_h.current_number_of_frames_in_rx_queue);
#endif

		callback_functions.got_rx_data(this, &rx_data);
		break;

	case SANG_STATUS_TDM_EVENT_AVAILABLE:
		callback_functions.got_TdmApiEvent(this, &rx_data);
		break;

	default:
		ERR_IFACE( "%s: Rx: Unknown Operation Status: %d\n", 
			device_name, pri->operation_status);
		break;
	}//switch()

	return 0;
}

///////////////////////////////////////////////////////////////////////

UCHAR sangoma_interface::transmit(TX_DATA_STRUCT *pTx)
{
	UCHAR rc = DoWriteCommand(hTx, pTx);
	if(rc){
		//error
		printf("DoWriteCommand() failed!! Check messages log.\n");
		return rc;
	}

	//check that frame was transmitted
	switch(pTx->api_header.operation_status)
	{
	case SANG_STATUS_SUCCESS:
		//printf("Frame queued for transmission.\n");
#if 0
		//Some useful information about API's internal transmit queue
		//is available after each successful transmission:
		printf("max_tx_queue_length: %d current_number_of_frames_in_tx_queue: %d\n",  
			pTx->api_header.tx_h.max_tx_queue_length,
			pTx->api_header.tx_h.current_number_of_frames_in_tx_queue);
#endif
		break;

	case SANG_STATUS_TX_TIMEOUT:
		//error
		//It means that tx buffers are constanly full,
		//usually because the line is disconnected
		printf("****** SANG_STATUS_TX_TIMEOUT ******\n");
		//Check messages log or look at statistics.
		break;
					
	case SANG_STATUS_TX_DATA_TOO_LONG:
		//Possible causes:
		//1. User specified pTx->api_header.data_length is too big.
		//2. Attempt to transmit data longer than the pre-configured maximum.
		//   Maximum length is set in 'Interface Properties',
		//   in the 'Device Manager'.
		//Check messages log.
		printf("****** SANG_STATUS_TX_DATA_TOO_LONG ******\n");
		break;
					
	case SANG_STATUS_TX_DATA_TOO_SHORT:
		//Possible cause:
		//1. Minimum is 1 byte.
		printf("****** SANG_STATUS_TX_DATA_TOO_SHORT ******\n");
		break;

	case SANG_STATUS_LINE_DISCONNECTED:
		//Front end monitoring is enabled and Line is
		//in disconnected state.
		//Check the T1/E1 line is in "Connected" state,
		//alse check the Alarms and the message log.
		printf("****** SANG_STATUS_LINE_DISCONNECTED ******\n");
		break;

	default:
		printf("Unknown return code (0x%X) on transmission!\n",
			pTx->api_header.operation_status);
		break;
	}//switch()
	return rc;
}

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
int sangoma_interface::start_ring_tone()
{
	DBG_IFACE( "%s:start_ring_tone()\n", device_name);
	return sangoma_tdm_enable_tone_events(WP_TDMAPI_EVENT_TONE_RING);
}

///////////////////////////////////////////////////////////////////////
int sangoma_interface::start_congestion_tone()
{
	DBG_IFACE( "%s:start_congestion_tone()\n", device_name);
	return sangoma_tdm_enable_tone_events(WP_TDMAPI_EVENT_TONE_CONGESTION);
}

///////////////////////////////////////////////////////////////////////
int sangoma_interface::start_busy_tone()
{
	DBG_IFACE( "%s:start_busy_tone()\n", device_name);
	return sangoma_tdm_enable_tone_events(WP_TDMAPI_EVENT_TONE_BUSY);
}

///////////////////////////////////////////////////////////////////////
int sangoma_interface::start_dial_tone()
{
	DBG_IFACE( "%s:start_dial_tone()\n", device_name);
	return sangoma_tdm_enable_tone_events(WP_TDMAPI_EVENT_TONE_DIAL);
}

///////////////////////////////////////////////////////////////////////
int sangoma_interface::stop_all_tones()
{
	DBG_IFACE( "%s:stop_all_tones()\n", device_name);
	return sangoma_tdm_disable_tone_events();
}

///////////////////////////////////////////////////////////////////////
int sangoma_interface::start_ringing_phone()
{
	DBG_IFACE( "%s:start_ringing_phone()\n", device_name);
	return sangoma_tdm_enable_ring_events();
}

int sangoma_interface::stop_ringing_phone()
{
	DBG_IFACE( "%s:stop_ringing_phone()\n", device_name);
	return sangoma_tdm_disable_ring_events();
}
///////////////////////////////////////////////////////////////////////
int sangoma_interface::fxo_go_off_hook()
{
	DBG_IFACE( "%s:fxo_go_off_hook()\n", device_name);

	if(wanif_conf_struct.sub_media != MOD_TYPE_FXO){
		ERR_IFACE( "%s: The 'Go Off Hook' command valid only for FXO!\n", device_name);
		return 1;
	}
	return sangoma_tdm_txsig_offhook();
}

int sangoma_interface::fxo_go_on_hook()
{
	DBG_IFACE( "%s:fxo_go_on_hook()\n", device_name);

	if(wanif_conf_struct.sub_media != MOD_TYPE_FXO){
		ERR_IFACE( "%s: The 'Go On Hook' command valid only for FXO!\n", device_name);
		return 1;
	}
	return sangoma_tdm_txsig_onhook();
}

//set a Telephony interface to it's default state
int sangoma_interface::reset_interface_state()
{
	DBG_IFACE( "%s:reset_interface_state()\n", device_name);

	switch(wanif_conf_struct.media)
	{
	case WAN_MEDIA_FXOFXS:
		switch(wanif_conf_struct.sub_media)
		{
		case MOD_TYPE_FXS:
			stop_all_tones();
			stop_ringing_phone();
			break;

		case MOD_TYPE_FXO:
			fxo_go_on_hook();
			break;
		}
		break;
	}
	return 0;
}


