//////////////////////////////////////////////////////////////////////
// sangoma_interface.h: interface for Sangoma API driver.
//
// Author	:	David Rokhvarg	<davidr@sangoma.com>
//////////////////////////////////////////////////////////////////////

//There is at least one communications Interface (represented by sangoma_interface class)
//running on a Port.

#ifndef SANGOMA_INTERFACE_H
#define SANGOMA_INTERFACE_H

#include <windows.h>
#include <winioctl.h>
#include <stdio.h>
#include <conio.h>
#include <stddef.h>		//for offsetof()

#define __WINDOWS__
#include <wanpipe_defines.h>	//for 'wan_udp_hdr_t'
#include <wanpipe_version.h>	//for DRIVER_VERSION structure
#include <sang_status_defines.h>//return codes
#include <sang_api.h>			//for IOCTL codes
#include <sdla_remora.h>		//Analog card
#include <wanpipe_cfg.h>		//for RBS commands, Media Types (WAN_MEDIA_NONE...)
#include <wanpipe_tdm_api_iface.h>	//for TDMV API
#include <sdla_aft_te1.h>		//for Wanpipe API
#include <wanrouter.h>			//for FE_CONNECTED/FE_DISCONNECTED

/*
 * FSK CallerID and DTMF detection note:
 *  Stelephony.dll expects input BitStream (aLaw or uLaw audio) from a SINGLE timeslot.
 *  It is always the case for Analog cards, but not for the Digital cards, where
 *  API may provide input from a SINGLE or from MULTIPLE timeslots.
 */
#define USE_STELEPHONY_API	0

#if USE_STELEPHONY_API
#include <StelephonyApi.h>
#endif

#include "sample_cpp.h"
#include "cthread.h"

/*	The OVERLAPPED_IO allows to use WaitForMultipleObjects() for receiving
	data and events on MULTIPLE Device Handles using SINGLE "Rx thread". */
//#define OVERLAPPED_IO

/**
  *@author David Rokhvarg
  */

class sangoma_interface : public CThread
{
	//////////////////////////////////////////////////////////////////
	HANDLE		hGeneralCommands;
	HANDLE		hRx;
	HANDLE		hTx;

	//////////////////////////////////////////////////////////////////
	//receive stuff
	ULONG		rx_frames_count;
	ULONG		rx_bytes_count;
	//for counting frames with CRC/Abort errors
	ULONG		bad_rx_frames_count;
	DWORD		dwRxThreadId; 

	//////////////////////////////////////////////////////////////////
	//transmit stuff
	HANDLE		hTxWaitHandle;
	ULONG		tx_bytes_count;
	ULONG		tx_frames_count;
	UCHAR		tx_test_byte;

	//////////////////////////////////////////////////////////////////
	//IOCTL management structures and variables
	wan_udp_hdr_t	wan_udp;
	int				protocol_cb_size;
	int 			wan_protocol;
	unsigned char	adapter_type;

	wanpipe_tdm_api_cmd_t tdm_api_cmd;

	void RxThreadFunc();
	int read_data();

	void PrintRxData(RX_DATA_STRUCT *pRx);
	void PrintTdmApiEvent(RX_DATA_STRUCT *pRx);
	void cleanup();

	wan_cmd_api_t			wanpipe_api_cmd;
	api_tx_hdr_t			*tx_hdr;

	int get_wan_config();
	int get_interface_configuration(if_cfg_t *wanif_conf_ptr);
	unsigned char get_wan_udphdr_data_byte(unsigned char off);
	unsigned char *get_wan_udphdr_data_ptr(unsigned char off);
	unsigned char set_wan_udphdr_data_byte(unsigned char off, unsigned char data);
	int get_fe_type(unsigned char* adapter_type);

	int wanpipe_api_ioctl(wan_cmd_api_t *api_cmd);

	USHORT	DoReadCommand(HANDLE drv, RX_DATA_STRUCT * pRx);
	UCHAR	DoWriteCommand(HANDLE drv, TX_DATA_STRUCT * pTx);
	UCHAR	DoSetIdleTxBufferCommand(HANDLE drv, TX_DATA_STRUCT *pTx);

	unsigned char wp_brt[256];
	void generate_bit_rev_table();

	char is_rbs_monitoring_enabled;
#if USE_STELEPHONY_API
	stelephony_callback_functions_t scf;
	void *stelObj;
#endif

	//////////////////////////////////////////////////////////////////
	//data
	char					terminate_tx_rx_threads;
	
	RX_DATA_STRUCT			rx_data;
	TX_DATA_STRUCT			tx_data;

	callback_functions_t	callback_functions;
	if_cfg_t				wanif_conf_struct;
	
	int WanpipeNumber, InterfaceNumber;

protected:
	virtual unsigned long threadFunction(void* parameter);	

public:
	char				device_name[DEV_NAME_LEN];	

	//////////////////////////////////////////////////////////////////
	//methods
	sangoma_interface(int wanpipe_number, int interface_number);
	~sangoma_interface();

	int DoManagementCommand(HANDLE drv, wan_udp_hdr_t* packet);
#define DO_COMMAND(wan_udp)	DoManagementCommand(hGeneralCommands, &wan_udp);

	int init(callback_functions_t *callback_functions_ptr);
	int run();
	int stop();

	void bit_swap_a_buffer(unsigned char *data, int len);

	void get_te1_56k_stat(void);
	void set_lb_modes(unsigned char type, unsigned char mode);
	
	int get_operational_stats(net_device_stats_t *stats);
	int flush_operational_stats (void);

	int enable_rbs_monitoring();
	char get_rbs(rbs_management_t *rbs_management_ptr);
	char set_rbs(rbs_management_t *rbs_management_ptr);

	int sangoma_tdm_enable_rbs_events(int polls_per_second);
	int sangoma_tdm_disable_rbs_events();

	void set_idle_tx_data_buffer(unsigned char idle_tx_test_byte);
	int get_open_handles_counter();

	//remove all data from API driver's transmit queue
	int flush_tx_buffers (void);

	HANDLE open_api_device(char *api_device_name);

	unsigned char get_adapter_type();
	unsigned int get_sub_media();

	void get_api_driver_version(PDRIVER_VERSION version);

	void get_card_customer_id(u8 *customer_id);

	UCHAR transmit(TX_DATA_STRUCT *pTx);

	USHORT receive(RX_DATA_STRUCT *pRx);

	//////////////////////////////////////////////////////////////////
	//TDM API calls
	int sangoma_tdm_enable_rxhook_events();
	int sangoma_tdm_disable_rxhook_events();

	/* DTMF Detection on A200 Analog card (SLIC) chip */
	int sangoma_tdm_enable_rm_dtmf_events();
	int sangoma_tdm_disable_rm_dtmf_events();

	/* DTMF Detection on Octasic chip */
	int sangoma_tdm_enable_dtmf_events(int channel);
	int sangoma_tdm_disable_dtmf_events(int channel);

	int sangoma_tdm_enable_ring_detect_events();
	int sangoma_tdm_disable_ring_detect_events(); 

	int sangoma_tdm_enable_ring_trip_detect_events(); 
	int sangoma_tdm_disable_ring_trip_detect_events();

	int sangoma_tdm_enable_ring_events();
	int sangoma_tdm_disable_ring_events();

	int sangoma_tdm_txsig_onhook();
	int sangoma_tdm_txsig_offhook();

	int sangoma_tdm_enable_tone_events(int tone_id);
	int sangoma_tdm_disable_tone_events();

	int sangoma_tdm_front_end_activate();
	int sangoma_tdm_front_end_deactivate();

	/* get current state of the line - is it Connected or Disconnected */
	int sangoma_tdm_get_front_end_status(unsigned char *status);
	//////////////////////////////////////////////////////////////////

	int tdmv_api_ioctl(wanpipe_tdm_api_cmd_t *api_cmd);

	int reset_interface_state();

	int start_ring_tone();
	int stop_ring_tone();

	int start_congestion_tone();
	int stop_congestion_tone();

	int start_busy_tone();
	int stop_busy_tone();

	int stop_all_tones();

	int start_dial_tone();
	int stop_dial_tone();

	int start_ringing_phone();
	int stop_ringing_phone();

	int fxo_go_off_hook();
	int fxo_go_on_hook();

#if defined(OVERLAPPED_IO)
	////////////////////////////////
	OVERLAPPED mOverlappedRx;
	OVERLAPPED mOverlappedTx;
	OVERLAPPED mOverlappedIdleTxBuffer;
	OVERLAPPED mOverlappedManagement;
	OVERLAPPED mOverlappedTdmvApiIoctl;
	////////////////////////////////
#endif
};

#endif//SANGOMA_INTERFACE_H

